# fixers.pl
#
# Perl subroutines to fix names etc.
# This file is used by some of the other perl scripts

# It is incorporated into those modules by the 'require' operator

# Fix capitalisation on names
sub FixName {
	local($IN) = @_;
	# Remove dots
	$IN =~ s/\\./ /;
	local(@Names) = split(/[ \\t\\n]+/,$IN);
	local($Res);
	foreach $Arg (@Names) {
		# Start by lower-casing the lot
		$Arg =~ tr/A-Z/a-z/;
		# Then Upper-case the first letter
		substr( $Arg,0,1 ) =~ tr/a-z/A-Z/;

		# Now look for Mc*
		substr( $Arg,2,1 ) =~ tr/a-z/A-Z/ if $Arg =~ /^Mc/;
		# Now look for O'* D'*
		substr( $Arg,2,1 ) =~ tr/a-z/A-Z/ if $Arg =~ /^.'/;
		# Now look for hyphenated names...
		$Hyphen = index($Arg, "-");
		substr( $Arg,$Hyphen+1,1 ) =~ tr/a-z/A-Z/ if $Hyphen >= 0;
	}
	$Res = join( ' ', @Names);
	# Deal with Dutch-style names
	$Res =~ s/Van Der /van der /;
	$Res =~ s/Van /van /;
	# Fix conjunctions for Department names
	$Res =~ s/ And / and /;
	$Res =~ s/ Of / of /;
	$Res =~ s/ The / the /;
	$Res =~ s/ & / and /;
	# Return the result
	$Res;
}

# Upper-case the initials and split them up

sub FixInitials {
	local($Arg) = @_;
	local($Res);
	$Arg =~ tr/a-z/A-Z/;
	$Res = join( ' ', split(/ */, $Arg) );
	# Hyphenated initials should probably be kept together
	$Res =~ s/ - /\-/g;
	$Res;
}

# Form commonName attribute lines in output

sub EmitCommonNames {
	local($Title, $Initials, $Forenames, $Surname) = @_;
	# RDN first

	printf( "cn= $Initials $Surname\\n" );

	# Another option is:
	# printf( "cn= $Title $Initials $Surname\\n" );

	# Other forms to assist matching
	local(@Names) = split(/[ \\t\\n]+/, $Forenames);
	printf( "cn= $Names[0] $Surname\\n" ) if $#Names >= 0;
}

# 'require' libraries must return true when loaded:
1;
