/* util.c */


/*****************************************************************************

11-May-94	str2AttrV() call replaced with str_at2AttrV() since attribute
		value string may have {T.61} headers.

03-Jun-94	skipHeader() added.

*****************************************************************************/
/*
 *				  NOTICE
 *
 *    Acquisition, use, and distribution of this module and related
 *    materials are subject to the restrictions of a license agreement.
 *    Consult the Preface in the User's Manual for the full terms of
 *    this agreement.
 *
 */


#include <signal.h>
#include <string.h>

#include <isode/quipu/util.h>
#include <isode/quipu/ds_search.h>
#include <isode/logger.h>

#include "types.h"
#include "util.h"
#include "namelist.h"
#include "bulkloadglobal.h"
#include "bulkloadattrnames.h"
#include "configuration.h"

extern LLog *bulkload_log;
extern int limitProblem;
extern int notAllReached;
extern int localAlarmTime;
extern int remoteAlarmTime;
extern int bulkloadLogLevel;
extern int alarmCount;
extern int bindTimeout;

extern	void	handleTimeout() ;

extern  struct	bulkLoadConfigStruct	bulkLoadConfig ;


int
listlen(lp)
struct namelist * lp;
{
int len;

  for (len = 0; lp != NULLLIST; lp = lp->next, len++) {};
  return len;
}


void startUnbindTimer()
{
void handleTimeout();

  (void) signal(SIGALRM, (VFP) handleTimeout);
  (void) alarm((unsigned)bindTimeout);
}

void stopUnbindTimer()
{
  (void) signal(SIGALRM, SIG_IGN);
  (void) alarm(0);
}

char *
copy_string(string)
char *string;
{
  char *new_string;

  if (string == NULLCP) return NULLCP;

  new_string = (char *) smalloc((strlen(string) + 1));
  (void) strcpy(new_string, string);

  return new_string;
}


static PS psdn = NULLPS;

char   *dn2pstr (dn)
DN	dn;
{
    char       *cp;

    if ((psdn == NULLPS
	    && ((psdn = ps_alloc (str_open)) == NULLPS))
		    || str_setup (psdn, NULLCP, SMALLBUFLEN, 0) == NOTOK) {
	if (psdn)
	    ps_free (psdn), psdn = NULLPS;

	return NULLCP;
    }

    dn_print (psdn, dn, RDNOUT);
    ps_print (psdn, " "); 
    *--psdn -> ps_ptr = NULL, psdn -> ps_cnt++;

    cp = psdn -> ps_base ;

    psdn -> ps_base = NULL, psdn -> ps_cnt = 0;
    psdn -> ps_ptr  = NULL, psdn -> ps_bufsiz = 0;

    return cp;
}


static PS psrdn = NULLPS;

char   *rdn2pstr (rdn)
RDN	rdn;
{
    char       *cp;

    if( (psrdn == NULLPS
	    && ((psrdn = ps_alloc (str_open)) == NULLPS))
		    || str_setup (psrdn, NULLCP, SMALLBUFLEN, 0) == NOTOK) {
	if (psrdn)
	    ps_free (psrdn), psrdn = NULLPS;

	return NULLCP;
    }

    rdn_print (psrdn, rdn, RDNOUT);
    ps_print (psrdn, " "); 
    *--psrdn -> ps_ptr = NULL, psrdn -> ps_cnt++;

    cp = psrdn -> ps_base ;

    psrdn -> ps_base = NULL, psrdn -> ps_cnt = 0;
    psrdn -> ps_ptr  = NULL, psrdn -> ps_bufsiz = 0;

    return cp;
}


static PS psat = NULLPS;

char *
AttrT2pstr( at )
AttributeType 	at ;
{
    char       *cp;

    if( (psat == NULLPS
	    && ((psat = ps_alloc (str_open)) == NULLPS))
		    || str_setup (psat, NULLCP, SMALLBUFLEN, 0) == NOTOK) {
	if (psat)
	    ps_free (psat), psat = NULLPS;

	return NULLCP;
    }

    AttrT_print ( psat, at, EDBOUT );
    ps_print (psat, " "); /* Stop truncation of last character ! */
    *--psat -> ps_ptr = NULL, psat -> ps_cnt++;

    cp = psat -> ps_base ;

    psat -> ps_base = NULL, psat -> ps_cnt = 0;
    psat -> ps_ptr  = NULL, psat -> ps_bufsiz = 0;

    return cp;
}


static PS psav = NULLPS;

char *
AttrV2pstr( av )
AttributeValue 	av ;
{
    char       *cp;

    if( (psav == NULLPS
	    && ((psav = ps_alloc (str_open)) == NULLPS))
		    || str_setup (psav, NULLCP, SMALLBUFLEN, 0) == NOTOK) {
	if (psav)
	    ps_free (psav), psav = NULLPS;

	return NULLCP;
    }

    AttrV_print ( psav, av, EDBOUT );
    ps_print (psav, " ");  /* Stop truncation of last character ! */
    *--psav -> ps_ptr = NULL, psav -> ps_cnt++;

    cp = psav -> ps_base ;

    psav -> ps_base = NULL, psav -> ps_cnt = 0;
    psav -> ps_ptr  = NULL, psav -> ps_bufsiz = 0;

    return cp;
}

#define 	BIGBUFLEN	10000

static PS psavs = NULLPS;

char *
avs2pstr( avs )
AV_Sequence avs ;
{
    char       *cp;

    if( (psavs == NULLPS
	    && ((psavs = ps_alloc (str_open)) == NULLPS))
		    || str_setup (psavs, NULLCP, BIGBUFLEN, 0) == NOTOK) {
	if (psavs)
	    ps_free (psavs), psavs = NULLPS;

	return NULLCP;
    }

    avs_print ( psavs, avs, EDBOUT );
    ps_print (psavs, " ");  /* Stop truncation of last character ! */
    *--psavs -> ps_ptr = NULL, psavs -> ps_cnt++;

    cp = psavs -> ps_base ;

    psavs -> ps_base = NULL, psavs -> ps_cnt = 0;
    psavs -> ps_ptr  = NULL, psavs -> ps_bufsiz = 0;

    return cp;
}


static PS psas = NULLPS;

char *
as2pstr( as )
Attr_Sequence as ;
{
    char       *cp;

    if( (psas == NULLPS
	    && ((psas = ps_alloc (str_open)) == NULLPS))
		    || str_setup (psas, NULLCP, BIGBUFLEN, 0) == NOTOK) {
	if (psas)
	    ps_free (psas), psas = NULLPS;

	return NULLCP;
    }

    as_print ( psas, as, EDBOUT );
    ps_print ( psas, " ");  /* Stop truncation of last character ! */
    *--psas -> ps_ptr = NULL, psas -> ps_cnt++;

    cp = psas -> ps_base ;

    psas -> ps_base = NULL, psas -> ps_cnt = 0;
    psas -> ps_ptr  = NULL, psas -> ps_bufsiz = 0;

    return cp;
}


/* 
 * Return an error message corresponding to the DSError structure
 * with each line preceded by a comment indicator (i.e. '#') .
 */ 

static PS pserror = NULLPS;

char *
error2pstr( errorp )
struct	DSError * errorp ;
{
char *	baseerrorcp;
char *	RtoLLinecp, * LtoRLinecp, * lastNcp ;
int 	newLineCount = 0 ;

    if( (pserror == NULLPS
	    && ((pserror = ps_alloc (str_open)) == NULLPS))
		    || str_setup (pserror, NULLCP, ERRORBUFSIZE, 0) == NOTOK) {
	if (pserror)
	    ps_free (pserror), pserror = NULLPS;

	return NULLCP;
    }

    ds_error( pserror, errorp ) ;
    ps_print( pserror, " " );  /* Stop truncation of last character ! */
    *--pserror -> ps_ptr = NULL, pserror -> ps_cnt++;

    LtoRLinecp = pserror -> ps_base ;

    /* Count the number of return characters.. */
    while( (LtoRLinecp = strchr( LtoRLinecp, '\n' )) != NULLCP )
    {
    	newLineCount++ ;

	LtoRLinecp++ ; /* skip the '\n' */
    }

    /* Create new storage.. */
    baseerrorcp = (char *)malloc( strlen( pserror -> ps_base ) + newLineCount + 5 ) ;

    if( baseerrorcp != NULLCP )
    {
	strcpy( baseerrorcp, pserror -> ps_base ) ;

	free( pserror -> ps_base ) ;

        pserror -> ps_base = NULL, pserror -> ps_cnt = 0;
 	pserror -> ps_ptr  = NULL, pserror -> ps_bufsiz = 0;

	/*
    	 * Insert a comment character after each new line
    	 */    

    	RtoLLinecp = baseerrorcp + strlen( baseerrorcp ) ; /* so that *RtoLLinecp = '\0' */

    	while( RtoLLinecp != NULLCP && newLineCount > 0 )
    	{
    	    /* Shift the current line.. */
	    do {

	    	*(RtoLLinecp + newLineCount) = *RtoLLinecp ;

	    } while( RtoLLinecp != baseerrorcp && *(--RtoLLinecp) != '\n' ) ;

	    /* Insert the comment character.. */
	    if( RtoLLinecp == baseerrorcp )
	    	*RtoLLinecp = '#' ; /* At start of line */
	    else
	    	*(RtoLLinecp + newLineCount) = '#' ;
	
	    newLineCount-- ;
    	}
    
        /* Remove the last '\n'.. */
	if( ( lastNcp = strrchr( baseerrorcp, '\n' )) != NULLCP )
	    *lastNcp = '\0' ;

#ifdef BLT_DEBUG
fprintf( stderr, "[[%s]]\n", baseerrorcp ); 
#endif
    }
    
    /* NOTE: the pserror stoage is now longer required and
       so will be cleared on the next iteration */
    
    return( baseerrorcp );
}


void
clearProblemFlags()
{
	limitProblem = notAllReached = FALSE;
}

void
setProblemFlags(sresult)
struct ds_search_result sresult;
{
	if ((sresult.CSR_limitproblem == LSR_SIZELIMITEXCEEDED) ||
	    (sresult.CSR_limitproblem == LSR_ADMINSIZEEXCEEDED))
		limitProblem = TRUE;
	if ((sresult.CSR_cr != NULLCONTINUATIONREF) ||
	    (sresult.CSR_limitproblem == LSR_TIMELIMITEXCEEDED))
		notAllReached = TRUE;
}

void
logSearchSuccess(outcome, objecttype, string, searchNumber, noMatches)
char * outcome;
char * objecttype;
char * string;
int searchNumber;
int noMatches;
{
char filterNumberString[20];

	if (bulkloadLogLevel > 1)
	{
		if (searchNumber == 0)
			(void) strcpy(filterNumberString, "explicit");
		else
			(void) sprintf(filterNumberString, "%d", searchNumber);
		
		(void) ll_log (bulkload_log, LLOG_NOTICE, NULLCP,
			"searchOutcome:%s:%s:%s:%s:%d", objecttype,
			outcome, string, filterNumberString, noMatches);
	}
}

void logListSuccess(outcome, objecttype, noMatches)
char * outcome;
char * objecttype;
int noMatches;
{
	if (bulkloadLogLevel > 1)
	{
		(void) ll_log (bulkload_log, LLOG_NOTICE, NULLCP,
			"listOutcome:%s:%s:%d", objecttype, outcome, noMatches);
	}
}

void logReadSuccess(outcome, objecttype)
char * outcome;
char * objecttype;
{
	if (bulkloadLogLevel > 1)
	{
		(void) ll_log (bulkload_log, LLOG_NOTICE, NULLCP,
			"readOutcome:%s:%s", objecttype, outcome);
	}
}


void
removeAllWhiteSpaces( tokencp )
char *	tokencp ;
{
char * 	fromcp = tokencp, * tocp = tokencp ;

	while( fromcp && *fromcp )
	{
	    if( !isspace( *fromcp ) )
	    	*tocp++ = *fromcp ;

	    ++fromcp ;
	}

	if( *fromcp == '\0' )
	    *tocp = '\0' ;

	return ;
}

/*
 * Remove white space characters (i.e. returns, new line, form feed or vertical tab).
 * Replace multiple spaces and tabs with a single one.
 * Remove line continuations "\\n".
 *
 * Return pointer is to malloc'ed storage.
 */

char *
strDupTidyStringRemLineConts( linecp )
char *	linecp ;
{
char *  malloccp, * tocp, * fromcp ;
char	suspectedLineContChar = '\0', lastCharSpace = '\0' ;

	while( isspace( *linecp ) )
	    linecp++ ; /* Skip leading whitespaces */

	fromcp = linecp ;

	tocp = malloccp = strdup( linecp ) ;

	if( tocp != NULL )
	{
	    while( *fromcp )
	    {
	    	if( isspace( *fromcp ) )
		{
		    if( *fromcp == ' ' || *fromcp == '\t' )
		    {
		    	lastCharSpace = *fromcp ;
		    }
		    else lastCharSpace = '\0' ;
			
		    suspectedLineContChar = '\0' ;
		}
		else if( *fromcp != LINECONTCHAR )
		{
 		    /* Note that in the case of the following string
 		     * 'A T \26 T', that you can have both lastCharSpace
 		     * and suspectedLineContChar at the same time
 		     * PJSG: 4/12/93
 		     */
 		    if( lastCharSpace )
 			*tocp++ = lastCharSpace ;
 
 		    lastCharSpace = '\0' ;
 		    
		    if( suspectedLineContChar )
		    {
		    	*tocp++ = suspectedLineContChar ; /* LINECONTCHAR */
			suspectedLineContChar = '\0' ;
		    }
		    
		    *tocp++  = *fromcp ;
		}
		else 
		{
		    suspectedLineContChar = LINECONTCHAR ;

		    /* Processing of "\n" notation added by J.C. 02/02/94 */
		    fromcp++ ;
		    if (*fromcp == 'n')
		    {
		    	/* Found "\n", copy newline to output line */
			*tocp++ = '\n';
			suspectedLineContChar = '\0' ;
		    }
		    else fromcp-- ; /* Unravel pointer advance */
		}
		
		fromcp++ ;
	    }

	    *tocp = '\0' ; /* Null terminate the string before leaving */
	}
	else fprintf( stderr, "Memory allocation failure in `strDupTidyStringRemLineConts()`\n" ) ;

	return( malloccp ) ;
}

/*
 * Replace any full stops within common name components with spaces.
 */

RDN 
str2bulkRdn( rdncp )
char *	rdncp ;
{
RDN	rdn = NULLRDN, currRdncomp, prevRdncomp = NULLRDN, nextRdncomp = NULLRDN ;
int	fullStopFound = FALSE ;
char *	avcp, * cnavcp, cnBuffer[ SMALLBUFLEN ] ;
static	AttributeType cnat = NULLAttrT ;

	if( bulkLoadConfig.replaceFullStopCommonNameDelimsWithSpaces == FALSE )
	    return( str2rdn( rdncp ) ) ;

	if( cnat == NULLAttrT )
	    cnat =  str2AttrT( BL_COMMON_NAME ) ;

	if( (currRdncomp = rdn = str2rdn( rdncp )) != NULLRDN )
	{
	    /* Check for any common name components.. */
	    while( currRdncomp != NULLRDN )
	    {
	    	if( AttrT_cmp( currRdncomp->rdn_at, cnat ) == 0 )
		{
		    fullStopFound = FALSE ;

		    if( (cnavcp = avcp = AttrV2pstr( &(currRdncomp->rdn_av) )) != NULLCP )
		    {
			/* Skip any leading {T.61} component.. */
			avcp = skipHeader( cnavcp, (int*)NULL ) ;
			
		    	/* Replace any full stops with a space.. */
		    	while( *avcp )
		    	{
		    	    if( *avcp == '.' )
			    {
			    	*avcp++ = ' ' ;

			    	fullStopFound = TRUE ;
			    }
			    else avcp++ ;
			}
		    }

		    if( fullStopFound == TRUE )
		    {
		    	/* Replace the common name RDN component.. */
			nextRdncomp = currRdncomp -> rdn_next ;

			currRdncomp -> rdn_next = NULLRDN ; /* to prevent any remaining components being freed */
			rdn_free( currRdncomp ) ;

			cnavcp = TidyString( cnavcp ) ; /* In case any double-spaces have been created */

			sprintf( cnBuffer, "%s= %s", BL_COMMON_NAME, cnavcp ) ;

			currRdncomp = str2rdn( cnBuffer ) ;

			if( prevRdncomp == NULLRDN )
			    rdn = currRdncomp ;
			else
			    prevRdncomp -> rdn_next = currRdncomp ;

			currRdncomp -> rdn_next = nextRdncomp ;
		    }

		    free( cnavcp ) ;
		}

		prevRdncomp = currRdncomp ;
	    	currRdncomp = currRdncomp -> rdn_next ;
	    }
	}

	return( rdn ) ;
}

/*
 * Compare the two strings whilst ignoring all white space characters..
 */
 
int
nonWspceStrsMatch( string1, string2 )
char *	string1, * string2 ;
{
int	matched = TRUE ;
char *  str1cp = string1, * str2cp = string2 ;
int	ptrIncremented, AnyMoreWspecChars = FALSE ;
	
	if( string1 == NULLCP || string2 == NULLCP )
	    return( FALSE ) ;

	while( *str1cp && *str2cp && matched == TRUE )
	{
	    ptrIncremented = FALSE ;
	    
	    if( isspace( *str1cp ) )
	    {
	    	str1cp++ ;
		ptrIncremented = TRUE ;
	    }
	    
	    if( isspace( *str2cp ) )
	    {
	    	str2cp++ ;
		ptrIncremented = TRUE ;
	    }


	    if( ptrIncremented == FALSE )
	    {
	    	if( *str1cp != *str2cp )
	    	    matched = FALSE ;

	    	else {
	    	    str1cp++, str2cp++ ;
	    	}
	    }
	}

	if( matched == TRUE )
	{
	    /* Check to see if there are any unread non-white
	       space characters in either string..            */

	    while( *str1cp && AnyMoreWspecChars == FALSE )
	    {
	        if( !isspace( *str1cp++ ) )
	   	    AnyMoreWspecChars = TRUE ;
	    }


	    while( *str2cp && AnyMoreWspecChars == FALSE )
	    {
	    	if( !isspace( *str2cp++ ) )
	    	    AnyMoreWspecChars = TRUE ;
	    }

	    if( AnyMoreWspecChars == TRUE )
	    	matched = FALSE ;
	}

	return( matched ) ;
}

/*
 * Check for a leading {T.61} string - if found the return points
 * points to the first character after it.
 */
char * 
skipHeader( avcp, t61_flagp )
char *	avcp ;
int  *	t61_flagp ;
{
char * retavcp = avcp ;

	/* Skip any leading {T.61} component.. */
	if( strlen( avcp ) >= 6 &&
	    *avcp == '{' && *(avcp + 1) == 'T' && *(avcp + 2) == '.' &&
	    *(avcp + 3) == '6' && *(avcp + 4) == '1' && *(avcp + 5) == '}' )
	{
	    if( t61_flagp != (int*)NULL )
		*t61_flagp = TRUE ;

	    retavcp += 6 ;
	}

	return retavcp ;
}

Attr_Sequence
findASbyAttrT( as, at )
Attr_Sequence	as ;
AttributeType	at ;
{
Attr_Sequence	curras = as, retas = NULLATTR ;

	while( curras != NULLATTR )
	{
	    if( AttrT_cmp( curras -> attr_type, at ) == 0 )
	    {
	    	retas = curras ;
	    	break ;
	    }
	
	    curras = curras -> attr_link ;
	}

	return( retas ) ;
}

AttributeValue 
findAttrVinAVS( avs, av )
AV_Sequence	avs ;
AttributeValue	av  ;
{
AV_Sequence	curravs = avs ;
AttributeValue	retav = NULLAttrV ;

	while( curravs != NULLAV )
	{
	    if( AttrV_cmp( &(curravs -> avseq_av), av ) == 0 )
	    {
	    	retav =  &(curravs -> avseq_av) ;
	    	break ;
	    }

	    curravs = curravs -> avseq_next ;
	}

	return( retav ) ;
}

/* 
 * Insert a new attribute value sequence node at the head
 * of the current sequence..
 * Note: the first parameter is a pointer to an AV_Sequence
 */

int
insertAVintoAVS( baseAVSp, av )
AV_Sequence   * baseAVSp ;
AttributeValue	av ;
{
AV_Sequence	buffavs, currAVS ;
int	    	valueAlreadyPresent = FALSE ;

	if( !baseAVSp )
	    return( FALSE ) ;

	currAVS = *baseAVSp ;

	/* First check that it is not there already.. */
	while( currAVS != NULLAV )
	{
	    if( AttrV_cmp( &(currAVS -> avseq_av), av ) == 0 )
	    {
	    	valueAlreadyPresent = TRUE ;
		break ;
	    }
	    
	    currAVS = currAVS -> avseq_next ;
	}

	if( valueAlreadyPresent == FALSE )
	{	
	    buffavs   = *baseAVSp ;
	    *baseAVSp = avs_comp_new( AttrV_cpy( av ) ) ;
	    (*baseAVSp) -> avseq_next = buffavs ;
	}

	return( valueAlreadyPresent ) ;
}

/*
 * Remove the attribute value sequence node that contains
 * the specified attribute value..
 * Note: the first parameter is a POINTER to an AV_Sequence (in case the first node
 * is to be removed.)
 */

int
removeAVfromAVS( baseAVSp, av )
AV_Sequence   * baseAVSp ;
AttributeValue  av ;
{
int	ret_val = NOTOK ;
AV_Sequence   *	currAVSp = baseAVSp, prevAVS = NULLAV, buffAVS ;
	
	if( !baseAVSp )
	    return( FALSE ) ;

	while( currAVSp && (*currAVSp) != NULLAV )
	{
	    if( AttrV_cmp( &((*currAVSp)->avseq_av), av ) == 0 )
	    {
		buffAVS = *currAVSp ;
	    
	    	/* Unlink the node and return its storage.. */
		if( prevAVS != NULLAV )
		    prevAVS -> avseq_next = (*currAVSp) -> avseq_next ;
		else
		    *currAVSp = (*currAVSp) -> avseq_next ;

	    	buffAVS -> avseq_next = NULLAV ;
		avs_free( buffAVS ) ;

		ret_val = OK ;
		
		break ;
	    }

	    prevAVS  = *currAVSp ;
	    currAVSp = &((*currAVSp) -> avseq_next) ;
	}
	
	return( ret_val ) ;
}

/*
 * Remove the attribute sequence node of the specified attribute type..
 * Note: the first parameter is a POINTER to an Attr_Sequence (in case the first node
 * is to be removed.)
 */

int
removeATfromAS(  baseASp, at )
Attr_Sequence   * baseASp ;
AttributeType	at ;
{
int	ret_val = NOTOK ;
Attr_Sequence   * currASp = baseASp, prevAS = NULLATTR, buffAS ;

	if( !baseASp )
	    return( FALSE ) ;

	while( currASp && (*currASp) != NULLATTR )
	{
	    if( AttrT_cmp( (*currASp)->attr_type, at ) == 0 )
	    {
		buffAS = *currASp ;
	    
	    	/* Unlink the node and return its storage.. */
		if( prevAS != NULLATTR )
		    prevAS -> attr_link = (*currASp) -> attr_link ;
		else
		    *currASp = (*currASp) -> attr_link ;

	    	buffAS -> attr_link = NULLATTR ;
		as_free( buffAS ) ;

		ret_val = OK ;
		
		break ;
	    }

	    prevAS  = *currASp ;
	    currASp = &((*currASp) -> attr_link) ;
	}
	
	return( ret_val ) ;
}


/*
 * Count the number of "avseqcomp" components within the AV_Sequence.
 */

int
countAVcomps( baseAVS )
AV_Sequence   baseAVS ;
{
AV_Sequence   currAVScompp = baseAVS ;
int 	numAVcomps  = 0 ;

	while( currAVScompp != NULLAV )
	{
	    currAVScompp = currAVScompp -> avseq_next ;

	    numAVcomps++ ;
	}

	return( numAVcomps ) ;
}

/*
 * Count the number of "dncomp" nodes within the given DN..
 */

int
countDNcomps( baseDN )
DN	baseDN ;
{
DN	currDNcompp = baseDN ;
int 	numDNcomps  = 0 ;

	while( currDNcompp != NULLDN )
	{
	    currDNcompp = currDNcompp -> dn_parent ;

	    numDNcomps++ ;
	}

	return( numDNcomps ) ;
}

/*
 * Search for the specified attribute type RDN component and
 * return a pointer to it.
 */

RDN
findAttrTypecpRdnComp( attrTypecp, updateRDN )
char *	attrTypecp ;
RDN 	updateRDN ;
{
RDN	currRdncomp = updateRDN ;
int	loopExit = FALSE ;

	while( currRdncomp != NULLRDN && loopExit == FALSE )
	{
	    if( strcmp( attr2name( currRdncomp->rdn_at, OIDPART ), attrTypecp ) == 0 )
		loopExit = TRUE ;
	    else
	    	currRdncomp = currRdncomp -> rdn_next ;
	}

	if( loopExit == TRUE )
	    return( currRdncomp ) ;
	else
	    return( NULLRDN ) ;
}


/*
 * Search for the specified attribute type RDN component and
 * return a pointer to it.
 */

RDN
findAttrTypeRdnComp( at, updateRDN )
AttributeType at ;
RDN 	updateRDN ;
{
RDN	currRdncomp = updateRDN ;
int	loopExit = FALSE ;

	while( currRdncomp != NULLRDN && loopExit == FALSE )
	{
	    if( AttrT_cmp( currRdncomp->rdn_at, at ) == 0 )
		loopExit = TRUE ;
	    else
	    	currRdncomp = currRdncomp -> rdn_next ;
	}

	if( loopExit == TRUE )
	    return( currRdncomp ) ;
	else
	    return( NULLRDN ) ;
}

/*
 * Return a pointer to the last RDN within a DN..
 */
RDN
findLastRDN( thisDN )
DN	thisDN ;
{
DN	currDN = thisDN, lastDN = NULLDN ;

	while( currDN )
	{
	    lastDN = currDN ;
	    
	    currDN = currDN -> dn_parent ;
	}

	if( lastDN != NULLDN )
	    return( lastDN -> dn_rdn ) ;
	else
	    return( NULLRDN ) ;
}

/*
 * Return a pointer to the next to last DN component within a DN..
 */
DN
findNextToLastDN( thisDN )
DN	thisDN ;
{
DN	currDN = thisDN, prevDN = NULLDN ;

	while( currDN )
	{
	    if( currDN -> dn_parent == NULLDN )
	    	break ;
	    else
	    {
	    	prevDN = currDN ;
	    	currDN = currDN -> dn_parent ;
	    }
	}

	return( prevDN ) ;
}


int
isCharInSet( thisChar, set )
char	thisChar, * set ;
{
int	inSet = FALSE ;
char * 	setcp = set ;	

#ifdef BLT_DEBUG
printf( "char %c string %s\n", thisChar, set ) ;
#endif
	while( *setcp )
	{
	    if( thisChar == *setcp++ )
	    {
	    	inSet = TRUE ;
		break ;
	    }
	}

	return( inSet ) ;
}


/*
 * If the specified attribute type and value are not present
 * in the Attribute Sequence then add them.
 */

void
appendAttributeToAttrSeq( attrTypecp, attrValuecp, baseas )
char *	attrTypecp, * attrValuecp ;
Attr_Sequence	baseas ;
{
Attr_Sequence	curras = baseas, prevas = NULLATTR ;
AV_Sequence	curravs, prevavs = NULLAV ;
AttributeType	at ;
AttributeValue	av ;
int	attributeAdded = FALSE ;

	if( (at = str2AttrT( attrTypecp )) != NULLAttrT &&
	    (av = str_at2AttrV( attrValuecp, at )) != NULLAttrV )
	{
	    /* Check for an occurrence of the specified attribute type.. */

	    while( attributeAdded == FALSE && curras != NULLATTR )
	    {
	    	if( AttrT_cmp( at, curras -> attr_type ) == 0 )
		{
		    /* Check for the specified attribute value.. */

		    curravs = curras -> attr_value ;
	    
	    	    while( attributeAdded == FALSE && curravs != NULLAV )
		    {
			if( AttrV_cmp( av, &(curravs -> avseq_av) ) == 0 )
			    attributeAdded = TRUE ; /* Attribute Value already present */
		    	else
			{
			    prevavs = curravs ;
			    curravs = curravs -> avseq_next ;
			}
			
		    } /* Attribute Value Sequence loop */

		    if( attributeAdded == FALSE )
		    {
			if( prevas != NULLATTR )
			{
			    /* Value is not already present for the Attribute Type so add it .. */
			    prevavs -> avseq_next = avs_comp_new( av ) ;
			    AttrT_free( at ) ;
			    attributeAdded = TRUE ;
			}
		    }
		}

	    	if( attributeAdded == FALSE )
		{
		    prevas = curras ;
		    curras = curras -> attr_link ;
		}
		
	    } /* Attr_Sequence loop */

	    if( attributeAdded == FALSE )
	    {
	    	/* No attributes of the required type have been found.. */

		if( prevas != NULLATTR )
		    prevas -> attr_link = as_comp_new( at, avs_comp_new( av ), NULLACL_INFO ) ;
		else
		{
		    AttrT_free( at ) ;
		    AttrV_free( av ) ;
		}
	    }
	}

	return ;
}


char * 
extractParentDNandReturnDNcp( entryDN )
DN	entryDN ;
{
char *	parentDNcp = NULLCP ;
DN	nextToLastDNcomp = NULLDN, buffDNcomp = NULLDN ;

	if( entryDN != NULLDN )
	{
	    if( (nextToLastDNcomp = findNextToLastDN( entryDN )) != NULLDN )
	    {
	    	buffDNcomp = nextToLastDNcomp -> dn_parent ;
		nextToLastDNcomp -> dn_parent = NULLDN ;

		parentDNcp = dn2pstr( entryDN ) ;

		nextToLastDNcomp -> dn_parent = buffDNcomp ;
	    }
	}

	return( parentDNcp ) ;
}



char 	commentHeader[] = { "# " } ;
char 	returnTail[]    = { "\n" } ;

/*
 * Returned error messages do not have comment headers on the first line
 * nor returns on the last.
 */
void
appendErrMsg( errorMsgStore, appendMsgcp, errorBufLen )
char 	errorMsgStore[] ;
char *	appendMsgcp ;
int	errorBufLen ;
{
int	storelen ;
char *	cp ;

	storelen = strlen( errorMsgStore ) ;

	if( ( storelen + strlen( appendMsgcp ) + 5 ) < errorBufLen )
	{
	    cp = errorMsgStore + storelen ;

	    if( storelen == 0 )
	    	sprintf( cp, "%s", appendMsgcp ) ;
	    else
	    	sprintf( cp, "%s%s%s", returnTail, commentHeader, appendMsgcp ) ;
	}

	return ;
}
